from clinicedc_constants import NO, TMG, YES
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models.signals import m2m_changed, post_delete, post_save
from django.dispatch.dispatcher import receiver
from django.utils import timezone

from edc_notification.models import Notification

from ..constants import AE_TMG_ACTION, DEATH_REPORT_TMG_ACTION
from ..utils import get_ae_model


@receiver(m2m_changed, weak=False, dispatch_uid="update_ae_notifications_for_tmg_group")
def update_ae_notifications_for_tmg_group(action, instance, **kwargs):
    if getattr(instance, "userprofile", None):
        try:
            tmg_ae_notification = Notification.objects.get(name=AE_TMG_ACTION)
        except ObjectDoesNotExist:
            pass
        else:
            try:
                with transaction.atomic():
                    instance.groups.get(name=TMG)
            except ObjectDoesNotExist:
                instance.userprofile.email_notifications.remove(tmg_ae_notification)
            else:
                instance.userprofile.email_notifications.add(tmg_ae_notification)


@receiver(post_save, weak=False, dispatch_uid="update_ae_initial_for_susar")
def update_ae_initial_for_susar(sender, instance, raw, update_fields, **kwargs):
    if not raw and not update_fields:
        try:
            ae_susar_model_cls = get_ae_model("AeSusar")
        except LookupError:
            pass
        else:
            if isinstance(instance, (ae_susar_model_cls,)) and getattr(
                instance.ae_initial, "susar", None
            ):
                if instance.submitted_datetime:
                    if instance.ae_initial.susar_reported != YES:
                        instance.ae_initial.susar = YES
                        instance.ae_initial.susar_reported = YES
                        instance.ae_initial.save(update_fields=["susar", "susar_reported"])
                elif instance.ae_initial.susar_reported != NO:
                    instance.ae_initial.susar = YES
                    instance.ae_initial.susar_reported = NO
                    instance.ae_initial.save(update_fields=["susar", "susar_reported"])


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_ae_initial_susar_reported",
)
def update_ae_initial_susar_reported(sender, instance, raw, update_fields, **kwargs):
    if not raw and not update_fields:
        try:
            ae_initial_model_cls = get_ae_model("AeInitial")
        except LookupError:
            pass
        else:
            if isinstance(instance, (ae_initial_model_cls,)) and getattr(
                instance, "susar", None
            ):
                ae_susar_model_cls = get_ae_model("AeSusar")
                if instance.susar == YES and instance.susar_reported == YES:
                    try:
                        with transaction.atomic():
                            ae_susar_model_cls.objects.get(ae_initial=instance)
                    except ObjectDoesNotExist:
                        ae_susar_model_cls.objects.create(
                            ae_initial=instance, submitted_datetime=timezone.now()
                        )


@receiver(post_delete, weak=False, dispatch_uid="post_delete_ae_susar")
def post_delete_ae_susar(instance, **kwargs):
    try:
        ae_susar_model_cls = get_ae_model("AeSusar")
    except LookupError:
        pass
    else:
        if (
            isinstance(instance, (ae_susar_model_cls,))
            and instance.ae_initial.susar == YES
            and instance.ae_initial.susar_reported != NO
        ):
            instance.ae_initial.susar_reported = NO
            instance.ae_initial.save()


@receiver(m2m_changed, weak=False, dispatch_uid="update_death_notifications_for_tmg_group")
def update_death_notifications_for_tmg_group(action, instance, **kwargs):
    if getattr(instance, "userprofile", None):
        try:
            tmg_death_notification = Notification.objects.get(name=DEATH_REPORT_TMG_ACTION)
        except ObjectDoesNotExist:
            pass
        else:
            try:
                instance.groups.get(name=TMG)
            except ObjectDoesNotExist:
                instance.userprofile.email_notifications.remove(tmg_death_notification)
            else:
                instance.userprofile.email_notifications.add(tmg_death_notification)



from django.db import models
from django.db.models import UniqueConstraint

from edc_model.models import BaseUuidModel


class PanelManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, name, lab_profile_name):
        return self.get(name=name, lab_profile_name=lab_profile_name)


class Panel(BaseUuidModel):
    name = models.CharField(max_length=50)

    display_name = models.CharField(max_length=50)

    lab_profile_name = models.CharField(max_length=50)

    objects = PanelManager()

    def __str__(self):
        return self.display_name or self.name

    def natural_key(self):
        return self.name, self.lab_profile_name

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Panel"
        verbose_name_plural = "Panels"
        constraints = (
            UniqueConstraint(
                fields=["name", "lab_profile_name"],
                name="%(app_label)s_%(class)s_name_uniq",
            ),
        )



from django import forms

from ...models import Receive, ReceiveItem


class ReceiveForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        if (
            getattr(self.instance, "id", None)
            and ReceiveItem.objects.filter(receive=self.instance).exists()
        ):
            raise forms.ValidationError("Receive record cannot be changed.")
        return cleaned_data

    class Meta:
        model = Receive
        fields = "__all__"
        help_text = {"receive_identifier": "(read-only)"}  # noqa: RUF012
        widgets = {  # noqa: RUF012
            "receive_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }


class ReceiveFormSuper(forms.ModelForm):
    class Meta:
        model = Receive
        fields = "__all__"
        help_text = {"receive_identifier": "(read-only)"}  # noqa: RUF012
        widgets = {  # noqa: RUF012
            "receive_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar.view_mixin import NavbarViewMixin


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = "edc_visit_schedule/home.html"
    navbar_name = "edc_visit_schedule"
    navbar_selected_item = "visit_schedule"



from django.contrib import messages
from django.utils.translation import gettext as _

from ...models import StockRequestItem


def delete_items_for_stock_request_action(modeladmin, request, queryset):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            _("Select one and only one item"),
        )
    else:
        stock_request = queryset.first()
        deleted = StockRequestItem.objects.filter(stock_request=stock_request).delete()
        messages.add_message(
            request,
            messages.SUCCESS,
            _("Delete %(deleted)s items for %(stock_request)s")
            % dict(deleted=deleted, stock_request=stock_request),
        )



from .aliquot_listboard_view import AliquotListboardView
from .manage_box_listboard_view import ManageBoxListboardView
from .manage_manifest_listboard_view import ManageManifestListboardView
from .manifest_listboard_view import ManifestListboardView
from .pack_listboard_view import PackListboardView
from .process_listboard_view import ProcessListboardView
from .receive_listboard_view import ReceiveListboardView
from .requisition_listboard_view import RequisitionListboardView
from .result_listboard_view import ResultListboardView
from .verify_box_listboard_view import VerifyBoxListboardView



import random
import string


class LabelData:
    def __init__(self):
        self.gender = random.choice(["M", "F"])  # nosec B311  # noqa: S311
        self.subject_identifier = "999-99-9999-9"  # nosec B311
        self.reference = "".join(
            random.choices(string.ascii_letters.upper() + "23456789", k=6)  # nosec B311  # noqa: S311
        )
        self.sid = "12345"
        self.site_name = "AMANA"
        self.pills_per_bottle = 128



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .exceptions import RegisteredSubjectError

if TYPE_CHECKING:
    from edc_registration.models import RegisteredSubject


class RegisteredSubjectDoesNotExist(Exception):  # noqa: N818
    pass


def get_registered_subject_model_name() -> str:
    return getattr(
        settings,
        "EDC_REGISTRATION_REGISTERED_SUBJECT_MODEL",
        "edc_registration.registeredsubject",
    )


def get_registered_subject_model_cls() -> type[RegisteredSubject]:
    return django_apps.get_model(get_registered_subject_model_name())


def get_registered_subject(
    subject_identifier, raise_exception: bool | None = None, **kwargs
) -> RegisteredSubject | None:
    opts = dict(subject_identifier=subject_identifier, **kwargs)
    try:
        registered_subject = get_registered_subject_model_cls().objects.get(**opts)
    except ObjectDoesNotExist:
        registered_subject = None
    if raise_exception and not registered_subject:
        # the subject consent usually creates the registered subject
        # instance. Check the model is declared with
        # UpdatesOrCreatesRegistrationModelMixin
        raise RegisteredSubjectDoesNotExist(
            "Unknown subject. "
            f"Searched `{get_registered_subject_model_cls()._meta.label_lower}`. "
            f"Got {opts}."
        )
    return registered_subject


def valid_subject_identifier_or_raise(
    subject_identifier: str, raise_exception: bool | None = None
) -> bool:
    if not re.match(ResearchProtocolConfig().subject_identifier_pattern, subject_identifier):
        if raise_exception:
            raise RegisteredSubjectError(
                f"Invalid subject identifier format. "
                f"Valid pattern is `{ResearchProtocolConfig().subject_identifier_pattern}`. "
                f"See `edc_protocol.ResearchProtocolConfig().subject_identifier_pattern`. "
                f"Got `{subject_identifier}`."
            )
        return False
    return True



from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlencode
from warnings import warn

from clinicedc_constants import NO, YES
from django import template
from django.contrib.admin.templatetags.admin_modify import (
    submit_row as django_submit_row,
)
from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch
from django_revision.revision import site_revision

from edc_model_admin.utils import get_next_url
from edc_protocol.research_protocol_config import ResearchProtocolConfig

if TYPE_CHECKING:
    from django.contrib.sites.models import Site
    from django.core.handlers.wsgi import WSGIRequest as Base

    class WSGIRequest(Base):
        site: Site


register = template.Library()


class EdcContextProcessorError(Exception):
    pass


def get_request_object(context):
    """Returns a request object or raises EdcContextProcessorError."""
    request = context.get("request")
    if not request:
        raise EdcContextProcessorError(
            "Request object not found in template context. "
            "Try enabling the context processor "
            "'django.template.context_processors.request'"
        )
    return request


def get_subject_identifier(context):
    """Returns the subject identifier."""
    request = get_request_object(context)
    subject_identifier = request.GET.get("subject_identifier")
    if not subject_identifier:
        try:
            subject_identifier = context["subject_identifier"]
        except KeyError:
            try:
                subject_identifier = context["original"].subject_identifier
            except (KeyError, AttributeError):
                subject_identifier = None
    return subject_identifier


def get_cancel_url(context, cancel_attr=None):
    """Returns the url for the Cancel button on the change_form."""
    request = get_request_object(context)
    cancel_url = request.GET.dict().get("cancel_url")
    if not cancel_url:
        cancel_querystring = request.GET.dict().get(cancel_attr or "cancel")
        if cancel_querystring:
            url = None
            kwargs = {}
            for pos, value in enumerate(cancel_querystring.split(",")):
                if pos == 0:
                    url = value
                else:
                    kwargs.update({value: request.GET.get(value)})
            try:
                cancel_url = reverse(url, kwargs=kwargs)
            except NoReverseMatch as e:
                warn(f"{e}. Got {cancel_url}.")
        else:
            cancel_url = get_next_url(request, warn_to_console=False)
            if not cancel_url:
                url = context["subject_dashboard_url"]
                kwargs = {"subject_identifier": get_subject_identifier(context)}
                try:
                    cancel_url = reverse(url, kwargs=kwargs)
                except NoReverseMatch:
                    cancel_url = None
    return cancel_url


@register.inclusion_tag("edc_model_admin/edc_submit_line.html", takes_context=True)
def edc_submit_row(
    context,
    cancel_url: str | None = None,
    cancel_url_kwargs: dict | None = None,
    cancel_url_querystring_data: dict | None = None,
    show_delete: bool | None = None,
):
    """Returns context to django_submit_row.

    Add to context in add_view or change_view.

    See also ModelAdminRedirectAllToChangelistMixin
    """
    request: WSGIRequest = get_request_object(context)
    model_site_id: int | None = getattr(getattr(context["original"], "site", None), "id", None)
    if model_site_id and request.site.id != model_site_id:
        context["has_add_permission"] = False
        context["has_change_permission"] = False
        context["has_delete_permission"] = False
    else:
        show_save = context.get("show_save")
        if "save_next" in context:
            context["save_next"] = show_save
    if "show_cancel" in context:
        if cancel_url:
            cancel_url = reverse(cancel_url, kwargs=(cancel_url_kwargs or {}))
            if cancel_url_querystring_data:
                cancel_url = f"{cancel_url}?{urlencode(cancel_url_querystring_data)}"
        context["cancel_url"] = cancel_url or get_cancel_url(context)
    if show_delete is False:
        context["show_delete"] = show_delete
    return django_submit_row(context)


@register.inclusion_tag("edc_model_admin/logout_row.html", takes_context=True)
def logout_row(context):
    return dict(
        perms=context.get("perms"),
        user=context.get("request").user,
        request=context.get("request"),
        site_url=context.get("site_url"),
    )


@register.inclusion_tag("edc_model_admin/edc_revision_line.html", takes_context=True)
def revision_row(context):
    return dict(
        copyright=context.get("copyright") or ResearchProtocolConfig().copyright,
        institution=context.get("institution") or ResearchProtocolConfig().institution,
        revision=context.get("revision") or site_revision.tag,
        disclaimer=context.get("disclaimer") or ResearchProtocolConfig().disclaimer,
    )


@register.inclusion_tag("edc_model_admin/edc_instructions.html", takes_context=True)
def instructions(context):
    return {"instructions": context.get("instructions")}


@register.inclusion_tag("edc_model_admin/edc_additional_instructions.html", takes_context=True)
def additional_instructions(context):
    return {
        "additional_instructions": context.get("additional_instructions"),
        "notification_instructions": context.get("notification_instructions"),
    }


@register.filter
def get_label_lower(model) -> str:
    """
    Returns label_lower for a model.
    """
    if model:
        return model._meta.label_lower
    return ""


@register.inclusion_tag("edc_model_admin/yes_no_coloring.html", takes_context=False)
def yes_no_coloring(value) -> dict:
    context = dict(value=value)
    if value == YES:
        context.update(color="green")
    elif value == NO:
        context.update(color="red")
    return context


@register.inclusion_tag("edc_model_admin/navbar_for_admin_templates.html", takes_context=True)
def show_navbar_for_admin_templates(context):
    return context


@register.inclusion_tag(
    "edc_model_admin/navbar_for_admin_templates_b3.html", takes_context=True
)
def show_navbar_for_admin_templates_b3(context):
    return context



mysql_view: str = """
    select *, uuid() as `id`, now() as `created` from (
        select site_id, username, report_model, min(accessed) as `first_accessed`,
               max(accessed) as `last_accessed`, count(*) as `access_count`
        from edc_qareports_qareportlog
        group by username, report_model, site_id
    ) as A
    order by username, report_model
    """

pg_view: str = """
    select *, get_random_uuid() as id, now() as created (
        select site_id, username, report_model, min(accessed) as first_accessed,
               max(accessed) as last_accessed, count(*) as access_count
        from edc_qareports_qareportlog
        group by username, report_model, site_id
    ) as A
    order by username, report_model
"""

sqlite3_view = """
SELECT *, lower(
    hex(randomblob(4)) || '-' || hex(randomblob(2)) || '-' || '4' ||
    substr(hex( randomblob(2)), 2) || '-' ||
    substr('AB89', 1 + (abs(random()) % 4) , 1)  ||
    substr(hex(randomblob(2)), 2) || '-' ||
    hex(randomblob(6))
  ) as id, datetime() as created from (
        select site_id, username, report_model, min(accessed) as first_accessed,
               max(accessed) as last_accessed, count(*) as access_count
        from edc_qareports_qareportlog
        group by username, report_model, site_id
    ) as A
    order by username, report_model
"""


def get_view_definition() -> dict:
    return {
        "django.db.backends.mysql": mysql_view,
        "django.db.backends.postgresql": pg_view,
        "django.db.backends.sqlite3": sqlite3_view,
    }



from __future__ import annotations

from django.apps import apps as django_apps
from django.db.models import QuerySet

from .model_to_dataframe import ModelToDataframe

__all__ = ["read_frame_edc"]


def read_frame_edc(
    model_or_queryset: QuerySet | str,
    *,
    drop_sys_columns: bool | None = None,
    drop_action_item_columns: bool | None = None,
    read_frame_verbose: bool | None = None,
):
    if not isinstance(model_or_queryset, QuerySet):
        qs = django_apps.get_model(model_or_queryset).objects.all()
    else:
        qs = model_or_queryset
    m = ModelToDataframe(
        queryset=qs,
        drop_sys_columns=drop_sys_columns,
        drop_action_item_columns=drop_action_item_columns,
        read_frame_verbose=read_frame_verbose,
    )
    if "site" not in m.dataframe.columns:
        m.dataframe["site"] = m.dataframe.site_id
    return m.dataframe



from __future__ import annotations

from collections.abc import Callable


class FC:
    """A simple class of the eligible criteria for a field.

    value: value expected IF ELIGIBLE
    msg: message if value is NOT met / ineligible
    ignore_if_missing: skip assessment if the field does not have a value

    if value is a callable it must return True/False, True means eligible.
    """

    def __init__(
        self,
        value: str | list | tuple | range | Callable[..., bool] | None = None,
        msg: str | None = None,
        ignore_if_missing: bool | None = False,
        missing_value: str | None = None,
    ):
        self.value = value
        self.msg = msg
        self.ignore_if_missing = ignore_if_missing
        self.missing_value = missing_value

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value, self.msg, self.ignore_if_missing})"



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from edc_model.models import BaseUuidModel
from edc_registration.models import RegisteredSubject

from ..exceptions import ValueBoundryError
from ..utils import get_grade_for_value, get_normal_data_or_raise
from .grading_exception import GradingException

if TYPE_CHECKING:
    from .grading_data import GradingData
    from .normal_data import NormalData


class ReferenceRangeCollection(BaseUuidModel):
    name = models.CharField(max_length=50, unique=True)

    grade1 = models.BooleanField(default=False)
    grade2 = models.BooleanField(default=False)
    grade3 = models.BooleanField(default=True)
    grade4 = models.BooleanField(default=True)

    def __str__(self):
        return self.name

    def grades(self, label: str) -> list[int]:
        return self.reportable_grades(label)

    def default_grades(self) -> list[int]:
        """Default grades considered for this collection.

        See also model GradingException.
        """
        return [i for i in range(1, 5) if getattr(self, f"grade{i}")]

    def reportable_grades(self, label: str) -> list[int]:
        if not label:
            raise ValueError("Unable to determine reportable grades. Label may not be None")
        try:
            grading_exception = GradingException.objects.get(
                reference_range_collection=self, label=label
            )
        except ObjectDoesNotExist:
            reportable_grades = self.default_grades()
        else:
            reportable_grades = grading_exception.grades
        return reportable_grades

    def get_grade(
        self,
        value: float | int | None = None,
        label: str | None = None,
        units: str | None = None,
        subject_identifier: str | None = None,
        report_datetime: datetime | None = None,
        gender: str | None = None,
        dob: date | None = None,
        age_units: str | None = None,
        site: Site | None = None,
    ) -> tuple[GradingData | None, str | None]:
        if subject_identifier:
            rs_obj = RegisteredSubject.objects.get(subject_identifier=subject_identifier)
            dob = rs_obj.dob
            gender = rs_obj.gender
        grading_data, eval_phrase = get_grade_for_value(
            reference_range_collection=self,
            value=value,
            label=label,
            units=units,
            gender=gender,
            dob=dob,
            report_datetime=report_datetime,
            age_units=age_units,
            site=site,
        )
        return grading_data, eval_phrase

    def is_normal(
        self,
        value: float | int | None = None,
        label: str | None = None,
        units: str | None = None,
        subject_identifier: str | None = None,
        report_datetime: datetime | None = None,
        gender: str | None = None,
        dob: date | None = None,
        age_units: str | None = None,
    ) -> tuple[bool, NormalData]:
        if subject_identifier:
            rs_obj = RegisteredSubject.objects.get(subject_identifier=subject_identifier)
            dob = rs_obj.dob
            gender = rs_obj.gender
        normal_data = get_normal_data_or_raise(
            reference_range_collection=self,
            label=label,
            units=units,
            gender=gender,
            dob=dob,
            report_datetime=report_datetime,
            age_units=age_units,
        )
        try:
            is_normal = normal_data.value_in_normal_range_or_raise(
                value=value,
                dob=dob,
                report_datetime=report_datetime,
                age_units=age_units,
            )
        except ValueBoundryError:
            is_normal = False
        return is_normal, normal_data

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Reference Range Collection"
        verbose_name_plural = "Reference Range Collections"



from dataclasses import dataclass, field
from string import Template

from clinicedc_constants import YES

from .crf_subquery import CrfSubquery


class RequisitionSubqueryError(Exception):
    pass


@dataclass(kw_only=True)
class RequisitionSubquery(CrfSubquery):
    """Generate a SELECT query returning requisitions where
    is_drawn=Yes for a given panel but results have not been captured
    in the result CRF.

    For example requisition exists for panel FBC but results_fbc
    CRF does not exist.
    """

    panel: str = None
    subjectrequisition_dbtable: str | None = None
    panel_dbtable: str | None = None
    requisition_id_field: str | None = None
    template: str = field(
        init=False,
        default=Template(
            "select req.subject_identifier, req.id as original_id, "  # nosec B608  # noqa: S608
            "req.subject_visit_id, req.report_datetime, req.site_id, v.visit_code, "
            "v.visit_code_sequence, "
            "v.schedule_name, req.modified, '${label_lower}' as label_lower, "
            "'${label}' as label, count(*) as records "
            "from ${subjectrequisition_dbtable} as req "
            "left join ${dbtable} as crf on req.id=crf.${requisition_id_field} "
            "left join ${subjectvisit_dbtable} as v on v.id=req.subject_visit_id "
            "${left_joins} "
            "left join ${panel_dbtable} as panel on req.panel_id=panel.id "
            f"where panel.name='${{panel}}' and req.is_drawn='{YES}' and crf.id is null "
            "group by req.id, req.subject_identifier, req.subject_visit_id, "
            "req.report_datetime, req.site_id, v.visit_code, v.visit_code_sequence, "
            "v.schedule_name, req.modified"
        ),
    )

    def __post_init__(self):
        # default where statement if not provided and have fld_name.
        if self.where is None and self.fld_name:
            self.where = f"crf.{self.fld_name} is null"
        if not self.label_lower:
            raise RequisitionSubqueryError("label_lower is required")
        if not self.subjectvisit_dbtable:
            self.subjectvisit_dbtable = f"{self.label_lower.split('.')[0]}_subjectvisit"
        if not self.subjectrequisition_dbtable:
            self.subjectrequisition_dbtable = (
                f"{self.label_lower.split('.')[0]}_subjectrequisition"
            )
        if not self.panel_dbtable:
            self.panel_dbtable = "edc_lab_panel"
        if not self.requisition_id_field:
            self.requisition_id_field = "requisition_id"



import urllib


class ListboardFilter:
    def __init__(
        self,
        name=None,
        label=None,
        lookup=None,
        exclude_filter=None,
        default=None,
        position=None,
    ):
        self.name = name
        self.label = label or name
        self.position = position or 0
        self.exclude_filter = exclude_filter
        if exclude_filter:
            self.attr = "e"
        else:
            self.attr = "f"
        self.lookup = lookup or {}
        self.default = default

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.name}, {self.label}, "
            f"exclude_filter={self.exclude_filter}, {self.default})"
        )

    @property
    def querystring(self):
        return urllib.parse.urlencode({self.attr: self.name})

    @property
    def lookup_options(self):
        lookup_options = {}
        for k, v in self.lookup.items():
            try:
                lookup_options.update({k: v()})
            except TypeError:
                lookup_options.update({k: v})
        return lookup_options
